#!/usr/bin/env python
# -*- coding: latin1 -*-

###	 EN :		Get info from a little ups with low details (NUT with generic driver, see genericups),
###				send email with some details and shutdown the machine.
###	 FR :		Rcupre l'tat du courant avec un onduleur (NUT avec le driver gnrique, voir genericups),
###				envoie un mail avec quelques dtails et etteint la bcane.
###	 LICENCE :	This software is under GNU General Public License version 2. See www.fsf.org
###	 AUTHOR :	Guillaume Savary <elrik@tele2.fr>

version = "1.7.1"

import os
import os.path
import sys
import string
import smtplib
import getopt
import re
import signal
from socket import *
from select import select
from email.MIMEText import MIMEText
from datetime import datetime
from time import sleep, time


class UPSSocket(socket):
	
	def __init__(self):
		socket.__init__(self, AF_INET, SOCK_STREAM)
		socket.setblocking(self, 00)

	def connect(self, host, port):
		try:
			socket.connect(self, (host, port))
		except Exception, e:
			if e[0] != 115:
				raise "CANNOT CONNECT ON UPS"
		
	def push(self, data):
		buffer = data
		while len(buffer) > 0:
			select([], [self], [self])
			try:
				sent = self.send(data)
			except Exception:
				raise "LOST UPS CONNECTION"
			if sent == 0:
				raise "LOST UPS CONNECTION"
			buffer = buffer[sent:]
			
	def get_data_until(self, terminators):
		if type(terminators) == type("string"):
			terminators = [terminators]
		buffer = ""
		found = False
		while found == False:
			select([self], [], [self])
			rcvd = self.recv(8192)
			if len(rcvd) < 1:
				raise "LOST UPS CONNECTION"
			buffer = buffer + rcvd
			for delimiter in terminators:
				if string.find(buffer,delimiter)  != -1:
					found = True
		return buffer

	def authenticate(self, username, password):
		self.push("USERNAME " + username + "\n")
		response = self.get_data_until("\n")
		
		if response[0:2] != "OK":
			raise "AUTHENTICATED FAILED"
		
		self.push("PASSWORD " + password + "\n")
		response = self.get_data_until("\n")
		
		if response[0:2] != "OK":
			raise "AUTHENTICATED FAILED"
		
	def logout(self):
		self.push("LOGOUT\n")
		self.close()

	
class ups_control:
	
	translation_ups={"OL":"On Line", "OB":"On Battery", "LB":"Low Battery", "ES":"ERROR UPS SCAN", "SD":"Shutdown Load", "CP":"Cable Power"}
	socket = None
	normal_time = 15
	stress_time = 5
	upsname = 'ups'
	global version
	
	def __init__(self, log, upsname, mailfrom, mailto, debug, command_shutdown, max_minutes_ups, max_continu):
		self.debug = debug
		self.log = log
		self.mailfrom = mailfrom
		self.mailto = mailto
		self.command_shutdown = command_shutdown
		self.max_minutes_ups = max_minutes_ups
		self.max_continu = max_continu
		self.upsname = upsname
		self.sec = 0

	def ups_connect(self, host, user='', pwd=''):
		self.socket = UPSSocket()
		self.socket.connect(host, 3493)
		select([self.socket], [self.socket], [self.socket])
		if user != '' and pwd != '':
			self.log.message('authenticated user %s, host %s\n' % (user, host))
			self.socket.authenticate(user, pwd)

	def get_ups_status(self):
		self.socket.push("GET VAR %s ups.status\n" % self.upsname)
		response =  self.socket.get_data_until(["\n","ERR"])
		# TODO error manage
		if response[-4:-2] in self.translation_ups.keys():
			self.ups = response[-4:-2]
		else:
			self.ups = "ES"
		
	def get_ups_info(self):
		self.log.log_date()
		self.socket.push("LIST VAR " + self.upsname + "\n")
		response =  self.socket.get_data_until(["END LIST VAR","ERR"])
		# TODO error manage
		lines = string.split(response,"\n");
		expression = r'VAR [^ \t\n\r\f\v]+ ([^ \t\n\r\f\v]+) "(.*)"'
		linelist = []
		for line in lines:
			matches = re.findall(expression,line)
			if len(matches) > 0:
				linelist = linelist + [matches[0]]
		for line in linelist:
			key, value = line
			if key in ('driver.name', 'driver.parameter.port', 'driver.version', 'ups.mfr', 'ups.model', 'ups.status'):
				self.log.log(key, value)
		control.log.log("ups_control version", version)

	def generate_mail(self):
		date=datetime.today()
		date_format=date.strftime("%d/%m/%Y - %H:%M")
		msg="Ups_control has detected interruption at %s during " % date_format
		if self.sec < 60:
			msg += "%d seconds." % self.sec
		else:
			msg += "%d minute(s) and %d seconds." % (self.sec/60, self.sec%60)

		msg+="\nups status : %s" % self.translation_ups[self.ups]
		if self.ups!='OL' and self.ups!='ES':
			msg+="\n\n*machine halt*"
		else:
			msg += "\n\n*machine is running*"
		email = MIMEText(msg)
		email['Subject']="ups_control low power detection"
		email['from']=self.mailfrom
		email['to']=self.mailto

		s = smtplib.SMTP()
		s.connect()
		s.sendmail(self.mailfrom, [self.mailto, self.mailfrom], email.as_string())
		self.log.log("send mail to", self.mailto)
		s.close()
			
	def shutdown(self):
		os.execl(self.command_shutdown)

	def run_loop(self):
		isZeTime(1)
		while 1:
			self.main()
			if self.max_continu:
				if self.ups == 'OL' or self.ups == 'ES':
					sleep(self.normal_time)
				else:
					sleep(self.stress_time)
			else:
				# what can we do here ?
				self.log.log("STOP!!", "stop trying to get ups info")
				self.generate_mail()
				sys.exit(0)

	def main(self):
		self.get_ups_status()
		if self.ups == 'OL' and self.sec == 0:
			self.max_continu = 3
		elif self.ups == 'OL' and self.sec > 0:
			self.generate_mail()
			self.sec = 0
		elif self.ups == 'OB' and self.sec <= self.max_minutes_ups * 60:
			self.sec += self.stress_time
		elif self.ups == 'LB' or (self.ups == 'OB' and self.sec > self.max_minutes_ups * 60):
			self.sec += self.stress_time
			self.generate_mail()
			self.sec = 0
			self.shutdown()
		else:
			self.max_continu-=1
			self.sec += self.normal_time
		if self.ups != 'OL' or self.sec != 0 or isZeTime(0):
			self.log.log_date()
			self.log.log('ups response', self.translation_ups[self.ups])
			

class Log:
	def __init__(self, _debug, _file_name="/var/log/ups_control.log"):
		self.debug=_debug
		try:
			self.file=open(_file_name, 'a')
		except:
			self.file = open(os.getenv('HOME')+'/ups_control.log', 'a')
			print "Can't write on /var/log/ups_control.log\nUse %s/ups_control.log" % os.getenv('HOME')
		
	def message(self, str):
		self.file.write('%s\n' % str)
		self.flush()
			
	def log(self, title, str):
		self.file.write(' %s %s %s\n' % (title, '.'*(24-len(title)), str))
		self.flush()

	def log_debug(self, title, str):
		if self.debug:
			self.log(title, str)

	def log_date(self):
		date=datetime.today()
		date_format=date.strftime("%d/%m/%Y - %H:%M:%S")
		self.message('-'*5 + ' ' + date_format + ' ' + '-'*5)
		
	def flush(self):
		self.file.flush()


def isZeTime(init):
	if init != 0:
		isZeTime.time = int(time())
	elif init == 0:
		now = int(time())
		if (now - isZeTime.time) >= 60*60:
			isZeTime.time = int(time())
			return True
	return False
		
	
def detach():
	pid = os.fork()
	if pid:
		try:
			file_run = open('/var/run/ups_control.pid', 'w')
			file_run.write('%d' % pid)
			file_run.close()
		except:
			print "Can't write in /var/run/ups_control.pid : daemon stop impossible"
		sys.exit(0)
	os.setsid()
	catch_signal()


def catch_signal():
	for sig in [ signal.SIGHUP, signal.SIGINT, signal.SIGQUIT, signal.SIGTERM ]:
		signal.signal(sig, signal_treatment)


def signal_treatment(sig, event):
	global log
	log.log_date()
	log.message('ups_control closed by signal %s\n' % sig)
	os.system('/bin/rm -f /var/run/ups_control.pid')
	sys.exit(0)


def host_treatment(str):
	index = str.find('@')
	if index>0:
		user, rest = str[:index], str[index+1:]
		index = rest.find(':')
		if index:
			host, pwd = rest[:index], rest[index+1:]
			return (host, user, pwd)
		else:
			usage()
	else:
		return (str, '', '')

	
def usage():
	"""
Ups_control check your NUT ups.

 usage : ups_control.py -t <mail> -n name [ -h user@host:pwd -f <mail> -s <command> -c <integer> -m <integer> --debug ]
         ups_control.py --stop
 -t : mail address to (your email adresse)
 -f : mail address from (your nut system adresse = mail_to by default)
 -m : max minutes before shutdown_command. Put 100 here if you prefer wait for the true shutdown signal
 -c : max number of errors we tolerate
 -s : what you want to do when shutdown event
 -h : user@host:pass - default = localhost without login. Exemple elrik@localhost:password
 --debug:  debug / verbose mode
 --stop: stop ups_control
"""
	print usage.__doc__
	sys.exit(-1)
		
	
if __name__ == "__main__":
	
	debug = 0
	command_shutdown = '/usr/bin/halt'
	mail_from = mail_to = ''
	upsname = ''
	host = 'localhost'
	max_continu = 3
	max_minutes_ups = 10
	stop = 0
	
	try:
		optlist, args = getopt.getopt(sys.argv[1:], 's:n:m:c:f:t:h:', ['debug', 'stop'])
	except getopt.GetoptError:
		usage()

	for opt, arg in optlist:
		if opt == '-s': command_shutdown = arg
		elif opt == '-m' : max_minutes_ups = arg
		elif opt == '-c' : max_continu = arg
		elif opt == '-f' : mail_from = arg
		elif opt == '-t' : mail_to = arg
		elif opt == '-n' : upsname = arg
		elif opt == '-h' : host = arg
		elif opt == '--debug' : debug = 1
		elif opt == '--stop' : stop = 1

	if stop == 1:
		print "try close ups_control ... ",
		try:
			pid=int(open('/var/run/ups_control.pid', 'r').read())
			os.system('/bin/kill -TERM %d' % pid)
			print "[  OK  ]\n -> pid %d detected" % pid
		except:
			print "[FAILED]"
			print "Can't read /var/run/ups_control.pid. Is ups_control running ?"
		else:
			sys.exit(0)
			
	if mail_to == '' or upsname == '':
		usage()
		
	if mail_from == '':
		mail_from = mail_to

	host, user, pwd = host_treatment(host)
	
	if debug :
		print host, ' ',  user, ' ', pwd
	else:
		detach()

	if stop == 0:
		print "try to start ups_control ... ",
		pid = -1
		try:
			pid=int(open('/var/run/ups_control.pid', 'r').read())
			print "[FAILED]\n -> pid %d detected. Is ups_control running ? " % pid
		except:
			pass
		else:
			if pid:
				sys.exit(-1)

	global log
	log = Log(debug)
	log.log_date()
	log.message("ups_control started\n")
	try:
		control = ups_control(log, upsname, mail_from, mail_to, debug, command_shutdown, max_minutes_ups, max_continu)
		control.ups_connect(host, user, pwd)
		control.get_ups_info()
		print "[  OK  ]"
	except:
		print "[FAILED]"
		sys.exit(-1)
	control.run_loop()
